/*
 * NN_throught_pass_kicker.cpp
 *
 *  Created on: 2017/07/25
 *      Author: mzmt
 */

#include "NN_through_pass_kicker.h"

using namespace rcsc;
using namespace std;

/*-------------------------------------------------------------------*/
/*! TODO README !!!!
 * パスを行う選手が実行する(ここ重要)、NNスルーパス
 * 計算量の問題があるので、
 * 毎サイクルにRione_libraryで1回だけ行う関数 createThroughtPassKick と
 * SampleFieldEvaluator内で、NNで出た値の参照のみ行う getThroughtPassKickPoint に分けた
 */

NN_kick &
NN_kick::instance()
{
    static NN_kick s_instance;
    return s_instance;
}

/*-------------------------------------------------------------------*/
/*
 * ここに各選手に向けて撃つべきスルーパスの地点を突っ込む
 */
std::vector< rcsc::Vector2D > NN_kick::throughPassPoint;

/*-------------------------------------------------------------------*/
/*!
 * スルーパスすべき座標を取得する
 * 毎サイクル実行されて、NNの計算したりするのはココ
 * 出力される座標は throughPassPoint に格納される
 * スルーパス要素が無い場合は、自分の座標をthroughPassPointに入れる例外処理
 */
void
NN_kick::createThroughtPassKick( const WorldModel & wm )
{
	// まずは消す
	throughPassPoint.clear();

	// ↓ 入力データ ↓

	// 自分の座標
	const int spos = convertPosToFieldNumber( wm.self().pos() );

	// 味方の座標 を ループ
	// その中で敵座標も取得し、NNを使って結果出力しpush_back、変な値は除く
	const PlayerPtrCont::const_iterator m_end = wm.teammatesFromSelf().end();
	for (PlayerPtrCont::const_iterator m = wm.teammatesFromSelf().begin(); m != m_end; ++m)
	{
		// 例外処理
		if( !(*m) ) continue;
		if( (*m)->isGhost() ) continue;
		if( (*m)->pos().x < wm.self().pos().x ) continue;

		// フィールドの座標番号取得
		const int mpos = convertPosToFieldNumber( (*m)->pos() );
		if ( mpos <= 0 ) continue;

		// 敵のスキマ座標
		Vector2D o_middle_pos = getOppMiddlePos( wm, (*m)->pos() );
		if ( o_middle_pos.absX() > 52.5
				|| o_middle_pos.absY() > 34.0)
		{
			continue;
		}

		// 敵のスキマ座標を番号に変換
		const int opos = convertPosToFieldNumber( o_middle_pos );

		// --------------------------------------------------------
		/* TODO:
		 * ここから NN を使う
		 *
		 * Input1 : spos ( int型、蹴る選手(自分)の座標番号 )
		 * Input2 : mpos ( int型、受け取る選手(ループ中の味方)の座標番号)
		 * Input3 : opos ( int型、mposに近い敵2人の中間地点の座標番号)
		 *
		 * Output : through_pass_pos ( 撃つべきスルーパス座標 )
		 *
		 */

		// 出力結果(座標番号)
		const int through_pass_pos = 0; // = NN(spos,mpos,opos); // XXX ここで計算 XXX

		if ( through_pass_pos <= 0
				|| through_pass_pos > 200 )
		{
			continue;
		}

		// その座標番号のマスの座標の中心あたりを返す
		Vector2D through_pass_point
			= Vector2D( (through_pass_pos / 10) * 5.25 + 2.6 ,	// x座標
						(through_pass_pos % 20) * 3.4 - 1.7 );	// y座標

		// 座標をpush_back
		throughPassPoint.push_back( through_pass_point );

	}

	// おわり か？
}

/*-------------------------------------------------------------------*/
/*!
 * SampleFieldEvaluator内で実際に使う関数
 * 引数に候補の地点を貰って、そこから一番近い NNで出た座標との距離を返す
 */
Vector2D
NN_kick::getNearestNNposFromPoint( const WorldModel & wm , const Vector2D state_ball_pos )
{
	double dist = 10000;
	Vector2D nnpos = Vector2D(0,0);

	for ( auto pos : throughPassPoint ){
		if ( dist > pos.dist( state_ball_pos ) ){
			dist = pos.dist( state_ball_pos );
			nnpos = pos;
		}
	}

	return nnpos;
}


/*-------------------------------------------------------------------*/
/*!
 * mate_posから近い敵2人 or 敵1人とサイドライン上の点 の中間地点を返す
 */
Vector2D
NN_kick::getOppMiddlePos( const WorldModel & wm, const Vector2D mpos )
{
	// サイドライン上のポイントをもとめる
	Vector2D side_point( mpos.x, mpos.absY() );
	if ( mpos.y < 0 )
		side_point.y *= -1.0;

	// その距離
	double side_dist = mpos.dist( side_point );

	// 欲しい2点の座標
	Vector2D fast = side_point;
	Vector2D second = side_point;

	// 敵ループ
	const PlayerPtrCont::const_iterator opps_end = wm.opponentsFromSelf().end();
	for (PlayerPtrCont::const_iterator o = wm.opponentsFromSelf().begin(); o != opps_end; ++o)
	{
		if( !(*o) ) continue;
		if( (*o)->isGhost() ) continue;

		// 一回目
		if ( fast == second ){
			if ( (*o)->pos().dist( fast ) < (*o)->pos().dist( second ) )
			{
				fast = (*o)->pos();
			}else{
				second = (*o)->pos();
			}
		}

		// それ以降
		if ( fast.dist((*o)->pos()) > mpos.dist((*o)->pos()) )
		{
			second = fast;
			fast = (*o)->pos();
		}
	}

	// 中間地点を返す
	return ( fast + second ) / 2;
}

/*-------------------------------------------------------------------*/
/*!
 * 引数の座標を、フィールド分割した番号に変換する
 * 例外は基本的に0
 */
int
NN_kick::convertPosToFieldNumber( const Vector2D point )
{
	// フィールドの外に出たら
	if ( point.absY() > 34.0
			|| point.x < 0
			|| point.x > 52.5 )
	{
//		printf("ERROR:: NN %d\n",__LINE__);
		return 0;
	}

	// 座標
	double posx = point.x;
	double posy = point.y + 34.0;

	// 縦横で何個目の整数か
	posx = int( posx / 5.25 );
	posy = int( posy / 3.40 ) + 1;

	// 20*10マスの何番にあるか
	int fieldNum = int(posx)*20 + int(posy);

	// 何か事故ったときの例外
	if( fieldNum < 0
			|| fieldNum > 200 )
	{
//		printf("ERROR:: NN %d\n",__LINE__);
		return 0;
	}

	return fieldNum;
}


