// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "basic_move_FW_center_forward.h"
#include "exp_DecemberRose.h"

#include "strategy.h"

#include "bhv_basic_tackle.h"
#include "bhv_smart_move.h"
#include "bhv_wander_move.h"

#include "Rione_library.h"
#include "NN_through_pass_receiver.h"


#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include "neck_offensive_intercept_neck.h"

#include "bhv_CFW_OppMark.h"

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!

 */
bool
BasicMove_CenterForward::execute( PlayerAgent * agent )
{
    dlog.addText( Logger::TEAM,
                  __FILE__": Bhv_CenterForwardMove" );
//    exp_DecemberRose(agent).execute();

    //-----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.8, 80.0 ).execute( agent ) )
    {
        return true;
    }

    const WorldModel & wm = agent->world();
	const ServerParam & SP = ServerParam::i();
    /*--------------------------------------------------------*/
    // chase ball
    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();
    int our_min = std::min( self_min, mate_min );
    int chase;// chase rate

	const PlayerObject * fast_opp = wm.interceptTable()->fastestOpponent();
	const int fast_opp_unum = ( fast_opp ) ? fast_opp->unum() : -1;

	const PlayerObject * fast_mate = wm.interceptTable()->fastestTeammate();
	const int fast_mate_unum = ( fast_mate ) ? fast_mate->unum() : -1;

	// インターセプトテーブルの表示デバッグ
	agent->debugClient().addMessage( "S%d[%d] M%d[%d] O%d[%d]",wm.self().unum(),self_min
															  ,fast_mate_unum, mate_min
															  ,fast_opp_unum, opp_min );

    if( self_min < opp_min || mate_min < opp_min )
    {
        chase = 2;
        if ( wm.existKickableOpponent() && wm.ball().pos().x > -10.0 )
            chase = 15;
    }
    else
    {
        if( wm.existKickableTeammate() )
        {
            chase = -5;
        }
        else
            chase = 10;
    }

    if ( ! wm.existKickableTeammate()
         && ( self_min <= chase
              || ( self_min <= mate_min
                   && self_min < opp_min + chase )
              )
         )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": intercept" );
        Body_Intercept().execute( agent );
        agent->setNeckAction( new Neck_OffensiveInterceptNeck() );

        return true;
    }

    //
    // ゴール前で味方がパス撃てるとき、自分はいい感じに相手のスキマに入る
    //
	if (Bhv_WanderMove().execute( agent )){
		return true;
	}

	// XXX テスト
	// スルーパスを受け取るための移動
	//
	if ( our_min + 1 < opp_min
			&& mate_min + 1 < self_min
			&& wm.lastKickerSide() == wm.ourSide()
			&& wm.ball().pos().x > 0
			&& wm.self().pos().x > 10 )
	{
		Vector2D kicker = wm.ball().inertiaPoint( mate_min );
		Vector2D recv_pos = NN_recv().createThroughtPassRecv( wm, kicker );
		if ( recv_pos != wm.self().pos()
				&& recv_pos.dist( wm.self().pos() ) < mate_min * 1.5 + 3.0 )
		{
		    if ( Body_GoToPoint( recv_pos, 0.5, Strategy::get_normal_dash_power( wm )
		                           ).execute( agent ) )
		    {
			    agent->debugClient().addMessage( "NN_recv Dash !!!" );
		    }else{
		        Body_TurnToBall().execute( agent );
			    agent->debugClient().addMessage( "NN_recv Turn !!!" );
		    }
		    agent->debugClient().setTarget( recv_pos );
		    if ( wm.existKickableOpponent()
		         && wm.ball().distFromSelf() < 18.0 )
		    {
		        agent->setNeckAction( new Neck_TurnToBall() );
		    }
		    else
		    {
		        agent->setNeckAction( new Neck_TurnToBallOrScan() );
		    }
		    return true;
		}
	}


    // テスト実装( ディフェンス時に発動 )
    if( Bhv_CFWOppMark( 6.5, 5.0 ).execute( agent ) ){
    	return true;
    }

    Vector2D target_point = Strategy::i().getPosition( wm.self().unum() );
    double dash_power = Strategy::get_normal_dash_power( wm );


    //
    // ボールがペナルティエリア内ならボールに向かって走る
    //
    if ( wm.ball().pos().x > 36 && wm.ball().pos().absY() < 20
    		&& wm.self().distFromBall() > 10.0 )
    {
        target_point = wm.ball().pos();
        dash_power = SP.maxDashPower();
    }

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    // オフェンスフォーメーション選択のデバック
    const char * offensive_state = Rione_Library().returnOffensiveState().c_str();
    agent->debugClient().addMessage( "CF form %.0f %s", dash_power, offensive_state );
    agent->debugClient().setTarget( target_point );
    agent->debugClient().addCircle( target_point, dist_thr );

    if ( ! Body_GoToPoint( target_point, dist_thr, dash_power
                           ).execute( agent ) )
    {
        Body_TurnToBall().execute( agent );
    }

    if ( wm.existKickableOpponent()
         && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new Neck_TurnToBallOrScan() );
    }

    return true;
}
