// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "tactical_NAME_basic_move_DF_side_back.h"
#include "../bhv_block_move_point.h"
#include "bhv_mark_move.h"
#include "../exp_DecemberRose.h"

#include "strategy.h"

#include "bhv_basic_tackle.h"
#include "Rione_library.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_scan_players.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>

#include "neck_offensive_intercept_neck.h"

#include <iomanip>

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!

 */
bool
Tactical_NAME_BasicMove_SideBack::execute( PlayerAgent * agent )
{
    dlog.addText( Logger::TEAM,
                  __FILE__": Bhv_BasicMove" );
//    DecemberRose(agent).execute();

    //-----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.9, 80.0 ).execute( agent ) )
    {
        return true;
    }

    const WorldModel & wm = agent->world();
    /*--------------------------------------------------------*/
    // chase ball
    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();

    int chase = 3;// chase rate
    if ( ! wm.existKickableTeammate()
         && ( ( self_min <= chase && self_min < mate_min && self_min <= opp_min ) // ボールが近い場合
              || ( self_min <= mate_min && self_min < opp_min + chase )          // ボールが遠い場合
              )
         )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": intercept" );
        Body_Intercept().execute( agent );
        agent->setNeckAction( new Neck_OffensiveInterceptNeck() );

        return true;
    }
    static bool flag = false;

    //スルーパスされたか？
    if(!flag && wm.ball().inertiaPoint(15).x + 4< wm.self().pos().x)
        flag = true;
    else if(flag && wm.ball().inertiaPoint(15).x + 4> wm.self().pos().x)
        flag = !flag;

    // マークする状況でマークできるならする
    if( Bhv_MarkMove().execute( agent ) )
    {
        return true;
    }

//   	if(Block_MovePoint().execute(agent))
//    	return true;

    Vector2D target_point = Strategy::i().getPosition( wm.self().unum() );
    const double nomal_dash_p = Strategy::get_normal_dash_power( wm );
    const double max_dash_p = ServerParam::i().MAX_DASH_POWER;
    double dash_power;

    if (wm.ball().pos().x < wm.self().pos().x
    		|| (wm.ball().pos().x < -0.0 && !wm.existKickableTeammate()))
    {
    	dash_power = max_dash_p;
    }
    else
    {
    	dash_power = nomal_dash_p;
    }

//    if (wm.ourDefenseLineX() + 5 < target_point.x
//        && wm.ball().pos().x - wm.ourDefenseLineX() < 10 && wm.lastKickerSide() != wm.ourSide()) {
//        target_point.x = wm.ourDefenseLineX() + 3;
//    }

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    double opp_ofence_x = wm.theirOffensePlayerLineX();
    //ボールの場所が-40以下で敵がけれる状態の時
//    if(wm.ball().pos().x > -40.0 && wm.existKickableOpponent()) {
//        //敵のオフェンスラインがターゲットポイントよりも後ろの場合．
//        if (opp_ofence_x < target_point.x) {
//            double dis = std::min(9.0, ServerParam::i().ourTeamGoalPos().dist(wm.ball().pos())/4.0);
//            agent->debugClient().addMessage("go_oppofenceline");
//            target_point.x = opp_ofence_x - dis;
//        }
//    }

    //敵がボールけってる時に少しボール側に詰める
    if( wm.lastKickerSide() != wm.ourSide() && wm.ball().pos().x < -30.0)
        target_point.y += (wm.ball().pos().y - target_point.y)/5;

    //スルーパスのフラグが立ったら．
    if(flag) {
        agent->debugClient().addMessage("(ToT)/~~~");
        target_point.x = wm.ball().inertiaPoint(15).x;
        dash_power = max_dash_p;
    }
    dlog.addText( Logger::TEAM,
                  __FILE__": Bhv_BasicMove target=(%.1f %.1f) dist_thr=%.2f",
                  target_point.x, target_point.y,
                  dist_thr );

    agent->debugClient().addMessage( "sidebackmove%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    agent->debugClient().addCircle( target_point, dist_thr );

    if ( ! Body_GoToPoint( target_point, dist_thr, dash_power
                           ).execute( agent ) )
    {
        Body_TurnToBall().execute( agent );
    }

    if ( wm.existKickableOpponent()
         && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new Neck_TurnToBallOrScan() );
    }

    return true;
}

double
Tactical_NAME_BasicMove_SideBack::dash_power_SideBack( rcsc::PlayerAgent * agent )
{
    const WorldModel & wm = agent->world();
    static bool s_recover_mode = false;

    if ( wm.self().staminaModel().capacityIsEmpty() )
    {
        return std::min( ServerParam::i().maxDashPower(),
                         wm.self().stamina() + wm.self().playerType().extraStamina() );
    }

    // スタミナ関連
    if ( wm.self().staminaModel().capacityIsEmpty() )
    {
        s_recover_mode = false;
    }
    else if ( wm.self().stamina() < ServerParam::i().staminaMax() * 0.5 )
    {
        s_recover_mode = true;
    }
    else if ( wm.self().stamina() > ServerParam::i().staminaMax() * 0.7 )
    {
        s_recover_mode = false;
    }

    /*--------------------------------------------------------*/
    double dash_power = ServerParam::i().maxDashPower();
    const double my_inc
        = wm.self().playerType().staminaIncMax()
        * wm.self().recovery();

    if ( wm.ourDefenseLineX() > wm.self().pos().x
         && wm.ball().pos().x < wm.ourDefenseLineX() + 20.0 )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": (get_normal_dash_power) correct DF line. keep max power" );
        // keep max power
        dash_power = ServerParam::i().maxDashPower();
    }
    else if ( s_recover_mode )
    {
        dash_power = my_inc - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;

        dlog.addText( Logger::TEAM,
                      __FILE__": (get_normal_dash_power) recovering" );
    }
    // 味方が蹴れるとき
    else if ( wm.existKickableTeammate()
              && wm.ball().distFromSelf() < 20.0 )
    {
        dash_power = std::min( my_inc * 1.1,
                               ServerParam::i().maxDashPower() );
        dlog.addText( Logger::TEAM,
                      __FILE__": (get_normal_dash_power) exist kickable teammate. dash_power=%.1f",
                      dash_power );
    }
    // ボールが自分より後ろ && 自分の近く && 相手ボール
    else if ( wm.ball().pos().x < wm.self().pos().x 
              && std::abs(wm.ball().pos().y - wm.self().pos().y) < 15
              && !Rione_Library::i().returnBallHavingTeam())
    {
        dash_power = ServerParam::i().maxDashPower();
    }
    // その他
    else
    {
        dash_power = std::min( my_inc * 1.7,
                               ServerParam::i().maxDashPower() );
        dlog.addText( Logger::TEAM,
                      __FILE__": (get_normal_dash_power) normal mode dash_power=%.1f",
                      dash_power );
    }

    return dash_power;
}
