// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "tactical_NAME_basic_move_GK_goalie.h"
#include "bhv_basic_move.h"
#include "bhv_basic_tackle.h"
#include "neck_goalie_turn_neck.h"
#include "bhv_goalie_chase_ball.h"
#include "strategy.h"
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_stop_dash.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/line_2d.h>
#include <rcsc/soccer_math.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!

 */
bool
Tactical_NAME_BasicMove_GoalieBasic::execute( PlayerAgent * agent )
{
	const rcsc::WorldModel & wm = agent->world();
     Vector2D base_point = getTargetPoint( agent );
     Vector2D move_point;

   /*  if(wm.ball().pos().x > 0 )
    {
    	 const Vector2D sefty_point = Strategy::i().getPosition(wm.self().unum());
    	 move_point = sefty_point;
    }else
    {*/
     move_point = base_point;
     //    }

     if( wm.ball().state().ghost_count_ > 0 )
     {
    	 //後ろをみる.
    	 agent->doTurn( 180.0 );
    	 rcsc::Neck_ScanField().execute( agent );
    	 agent->debugClient().addMessage("ballLost");
    	 //     basicMove( agent );
    	 return true;
     }

     if( move_point.absY() > 7.0 ){
    	 move_point.y *= 0.5;
     }

     dlog.addText( Logger::TEAM,
    		 __FILE__": Bhv_GoalieBasicMove. move_point(%.2f %.2f)",
    		 move_point.x, move_point.y );

     //////////////////////////////////////////////////////////////////
    // tackle
    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": tackle" );
        return true;
    }

    /////////////////////////////////////////////////////////////
    //----------------------------------------------------------
    if ( doPrepareDeepCross( agent, move_point ) )
    {
        // face to opponent side to wait the opponent last cross pass
    	 agent->debugClient().addMessage( "doPrepareDeepCross!" );
        return true;
    }
    //----------------------------------------------------------
    // check distance to the move target point
    // if already there, try to stop
    if ( doStopAtMovePoint( agent, move_point ) )
    {
        // execute stop action
    	agent->debugClient().addMessage( "doStopAtMovePoint!" );
        return true;
    }
    //----------------------------------------------------------
    // check whether ball is in very dangerous state
    if ( doMoveForDangerousState( agent, move_point ) )
    {
        // execute emergency action
    	agent->debugClient().addMessage( "doMoveForDangerousState!" );
        return true;
    }
    //----------------------------------------------------------
    // check & correct X difference
    if ( doCorrectX( agent, move_point ) )
    {
        // execute x-pos adjustment action
    	agent->debugClient().addMessage( "doCorrectX!" );
        return true;
    }

    //----------------------------------------------------------
    if ( doGoToMovePoint( agent, move_point ) )
    {
    	// mainly execute Y-adjustment if body direction is OK. -> only dash
    	// if body direction is not good, nomal go to action is done.
    	agent->debugClient().addMessage( "doGoToMovePoint!" );
    	return true;
    }

    //----------------------------------------------------------
    if ( doCorrectBodyDir( agent, move_point, true ) ) // consider opp
    {
    	// exeucte turn
    	agent->debugClient().addMessage( "doCorrectBodyDir!" );
        return true;
    }


    //----------------------------------------------------------
    // change my body angle to desired angle
    if ( doCorrectBodyDir( agent, move_point, false ) ) // not consider opp
    {
    	agent->debugClient().addMessage( "doCorrectBodyDir!" );
        return true;
    }

    dlog.addText( Logger::TEAM,
                  __FILE__": only look ball" );
    agent->debugClient().addMessage( "OnlyTurnNeck8!!!!!" );

   // agent->doTurn( 0.0 );

    rcsc::AngleDeg angle = ( move_point - ( wm.self().pos() + wm.self().vel() ) ).th() - wm.self().body();
    agent->doTurn( angle );

   agent->setNeckAction( new rcsc::Neck_TurnToBall());

    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
Vector2D
Tactical_NAME_BasicMove_GoalieBasic::getTargetPoint( PlayerAgent * agent )
{//yota
	//キーパーのベースポジション
	//ケリ出す位置にいる
	 //const rcsc::ServerParam & sp = rcsc::ServerParam::i();
	const double base_move_x = -51.5;
    //const double danger_move_x = -51.2;
    const WorldModel & wm = agent->world();
    int ball_reach_step = 0;
    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();
    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
    const rcsc::PlayerPtrCont & opps_near_ball = wm.opponentsFromBall();
    const rcsc::PlayerPtrCont & ours = wm.teammatesFromSelf();
    static const Rect2D our_penalty( Vector2D( -ServerParam::i().pitchHalfLength(),
  	    			   -ServerParam::i().penaltyAreaHalfWidth() + 1.0 ),
  	    			   Size2D( ServerParam::i().penaltyAreaLength() - 1.0,
  	    					   	   ServerParam::i().penaltyAreaWidth() - 2.0 ) );
		Vector2D nearest_opp =  Vector2D(0.0,0.0);
  		Vector2D nearest_our =  Vector2D(0.0,0.0);
  		Vector2D nearestball_opp =  Vector2D(0.0,0.0);

    for( rcsc::PlayerPtrCont::const_iterator it = ours.begin();
	  	         it != ours.end();
	  	         ++it )
    {
    	const rcsc::PlayerObject * q = * it;
    	if ( ! q ) continue;
    	else{
    		nearest_our = q->pos();
    		break;
    	}
    }
    for( rcsc::PlayerPtrCont::const_iterator it = opps_near_ball.begin();
   	  	         it != opps_near_ball.end();
   	  	         ++it )
       {
       	const rcsc::PlayerObject * q = * it;
       	if ( ! q ) continue;
       	else{
       		nearestball_opp = q->pos();
       		break;
       	}
       }



    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( mate_min,
                        opp_min );
    }
    //base_posはballの予測地点がだせる。
    const Vector2D base_pos = wm.ball().inertiaPoint( ball_reach_step );

      //コーナーキック
	  if( wm.gameMode().side() != wm.ourSide()
	        && wm.gameMode().type() == rcsc::GameMode::CornerKick_ )
	    {		//いる場所指定して移動
	        const rcsc::Vector2D move_point( -52.2, 6.6 * ( wm.ball().pos().y / wm.ball().pos().absY() ) );
	        return move_point;
	    }
	  else if(( wm.gameMode().type() == rcsc::GameMode::KickIn_
			  || wm.gameMode().type() == rcsc::GameMode::OffSide_ )&& wm.ball().pos().x < 0)

	  {
		  //サイド 意味ある？
		  if(wm.ball().pos().x < -30.0 && wm.ball().pos().absY() > 32.0 )
		  {
			  //printf("kick");
			  agent->debugClient().addMessage( "kick" );
			  const rcsc::Vector2D move_point(-52.2, 6.5 * ( wm.ball().pos().y / wm.ball().pos().absY()));
			  return move_point;
		  }
		  else
		  {
			  const rcsc::Vector2D move_point(-52.0,0);
			  return move_point;
		  }
	  }
	  //フリーキック
	  else if(wm.gameMode().type() == rcsc::GameMode::FreeKick_||wm.gameMode().type() == rcsc::GameMode::FoulCharge_)
	  {

		  if(wm.ball().pos().x < -35.0 &&wm.ball().pos().absY() < 7.0){
			  const rcsc::Vector2D move_point( -52.2,0);
			  agent->debugClient().addMessage( "FreeKick" );
			  return move_point;
		  }
		  else if(wm.ball().pos().x < -35.0 &&wm.ball().pos().absY() > 9.0)
		  {

			  rcsc::Vector2D move_point( -52.0, 6.5 * ( wm.ball().pos().y / wm.ball().pos().absY() ) );
			  agent->debugClient().addMessage( "wm_ball -35 outside " );
			  return move_point;
		  }
		  else if(wm.gameMode().side() != wm.ourSide())
		  {
			  agent->debugClient().addMessage( "Oppside" );
			  //  printf("Opp");
			  if(nearest_opp.y > 6.5 )
			  {
				  nearest_opp.y = 6.5;
			  }else if(nearest_opp.y < -6.5)
			  {
				  nearest_opp.y = -6.5;
			  }
			  Vector2D Mark_point = rcsc::Line2D::intersection(rcsc::Line2D( rcsc::Vector2D( -52.0, -9.0 ),rcsc::Vector2D( -52.0, 9.0 )),
					  rcsc::Line2D( Vector2D(-53.0,nearest_opp.y),nearest_opp));
			  return Mark_point;
		  }
		  else
		  {
			  // printf("foul¥n");
			  agent->debugClient().addMessage( "foul" );
			  const rcsc::Vector2D move_point(-52.0,0);
			  return move_point;
		  }
	  }
	  else if( wm.gameMode().side() == wm.ourSide()
			  && wm.gameMode().type() == rcsc::GameMode::GoalKick_ )
	  {
		  const rcsc::Vector2D move_point( -51.0, 0.0 );
		  agent->debugClient().addMessage( "Goal_kick" );
		  return move_point;
	  }
	  else if( wm.gameMode().type() == rcsc::GameMode::KickOff_ )
	  {
		  const rcsc::Vector2D move_point( -40.0, 0.0 );
		  agent->debugClient().addMessage( "kick_off" );
		  return move_point;
	  }//ゴールに入りそうなときの処理
	  //これにより追いかけるような動作である
	  else if( rcsc::Segment2D( wm.ball().pos(), wm.ball().inertiaFinalPoint() ).existIntersection(   //Segment2D:線分
			  rcsc::Segment2D( rcsc::Vector2D( -52.0, -9.0 ), rcsc::Vector2D( -52.0, 9.0 ) ) ) )
	  {
		  //敵エージェントが近くにいるならば
		  //ゴール前で待つのポジションでまつ
		  agent->debugClient().addMessage("shootM");

		  //一番近くの敵エージェントがペナルティにいるとき
		  //ゴール前に線引いて，ボールの最終地点とボールの蹴られる場所に移動する
		  if(  wm.ball().inertiaPoint( self_min ).x > -51.0  )
		  {
			  Bhv_GoalieChaseBall().execute( agent );
			  agent->debugClient().addMessage("shootM(1)");
			  return wm.ball().inertiaTravel( self_min );
		  }
		  //nearest_oppがどこにいるかで
		  else if(our_penalty.contains(nearest_opp))
		  {
			  agent->debugClient().addMessage( "!!!!!!!!!!!!!" );
			  rcsc::Vector2D move_point(-52.0,0);
			  rcsc::Vector2D goal_1( -52.5 , ServerParam::i().goalHalfWidth() );
			  rcsc::Vector2D goal_2( -52.5 , 0 );
			  //ボールの位置、ゴールの中心、ゴールポストで三角形を2つ作る
			  Triangle2D triangle_right = rcsc::Triangle2D( wm.ball().pos() , goal_1 , goal_2);
			  Triangle2D triangle_left = rcsc::Triangle2D( wm.ball().pos() , goal_1 , goal_2);
			  int mate_count_right=0, mate_count_left=0;

			  const PlayerPtrCont::const_iterator mates_end = wm.teammatesFromSelf().end();
			  for (PlayerPtrCont::const_iterator mate = wm.teammatesFromSelf().begin();
					  mate != mates_end; ++mate)
			  {
				  if( triangle_right.contains( (*mate)->pos() )){
					  mate_count_right++;
				  }
				  else if( triangle_left.contains( (*mate)->pos() ) ){
					  mate_count_left++;
				  }
			  }

			  if( mate_count_right > mate_count_left  ){
				  move_point.y -= 2.0;
			  }
			  else if( mate_count_left > mate_count_right  ){
				  move_point.y += 2.0;
			  }

			  return move_point;
		  }

		  //慣性がかかったときにとまるボールの位置が52以上だった場合ボール追いかける
		  else
		  {/*----------------------------------------取りに行くのにどれくらいのサイクル数かをみる------------------------------------------------------------*/
			  rcsc::Vector2D move_point = rcsc::Line2D( wm.ball().pos(), wm.ball().inertiaFinalPoint() ).intersection(
					  rcsc::Line2D( rcsc::Vector2D( -52.5, -9.0 ), rcsc::Vector2D( -52.5, 9.0 ) ) );
			  //int cycle = 1;
			  for( int i = 1; i < 15; i++ )
			  {
				  //iサイクル後のボールの位置とキーパの位置の差が、move_pointとキーパの位置の差より小さいとき→ボールに追いつける
				  if( wm.ball().inertiaPoint( i ).dist( wm.self().pos() ) < move_point.dist( wm.self().pos() ) )
				  {
					  Line2D expect_ball_line( wm.ball().inertiaPoint(i) , wm.ball().pos() ) ;
					  move_point.y = expect_ball_line.getY( move_point.x );
					  // cycle = i;
				  }

			  }
			  agent->debugClient().addMessage("shootM(3)");
			  return move_point;
		  }
	  }
	  /*---------------------------------------------------------------------------------------------------*/
	  //
	  else if (base_pos.x < -ServerParam::i().pitchHalfLength() + 8.0
			  && base_pos.absY() > ServerParam::i().goalHalfWidth() + 5.0
			  && mate_min < self_min)
	  {
		  Vector2D move_point( base_move_x, ServerParam::i().goalHalfWidth() - 0.1 );
		  //このやり方で上下逆にできる
		  if ( base_pos.y < -0.5 )
		  {
			  move_point.y *= -1.0;
		  }

		  dlog.addText( Logger::TEAM,
				  __FILE__": getTarget. target is goal pole" );
		  agent->debugClient().addMessage( "Pos(1)" );
		  return move_point;
	  }
	  //取りに行く動作
	  if( self_min <= opp_min
			  && self_min < mate_min){
		  return wm.ball().inertiaPoint( wm.interceptTable()->selfReachCycle());
	  }
	  else if( base_pos.x < -22.0 && base_pos.absY() > 20.0){
		  doMoveForDangerousState(agent,Vector2D(-52.0,6.5 * ( wm.ball().pos().y / wm.ball().pos().absY())));
		  agent->debugClient().addMessage( "Pos(2)" );
		  return Vector2D(-52.0,6.0 * ( wm.ball().pos().y / wm.ball().pos().absY()));
	  }
	  else if( base_pos.x < -15.0 && base_pos.absY() < 20.0 )
	  {
		  Vector2D danger_point = rcsc::Line2D::intersection(rcsc::Line2D( rcsc::Vector2D( -51.5, -9.0 ),rcsc::Vector2D( -52.5, 9.0 )),
				  rcsc::Line2D( rcsc::Vector2D( -55.0, 0.0 ),wm.ball().pos()));
		  float buf = 1.2;
		  if(danger_point.y < 0){
			  buf *= -1.0;
		  }else if(base_pos.y > -0.5&&base_pos.y < 0.5)
		  {
			  buf = 0.0;
		  }
		  danger_point.y +=buf;
		  if(danger_point.y > 6.5)
		  {
			  danger_point.y = 6.5;
		  }
		  else if(danger_point.y < -6.5)
		  {
			  danger_point.y = -6.5;
		  }
		  doMoveForDangerousState(agent,Vector2D (-51.5,danger_point.y ) );
		  agent->debugClient().addMessage( "Pos(3)" );
		  return Vector2D(-51.5,danger_point.y);
	  }else
	  {
		  //ここにパスカット入れる
		  /*
	//	rcsc::Vector2D move_point = rcsc::Line2D( wm.ball().pos(), wm.ball().inertiaFinalPoint() ).intersection(
	    	 const PlayerObject* our_5 = Bhv_GoalieBasicMove::teammate(5,agent);
	    	 		 const PlayerObject* our_4 = Bhv_GoalieBasicMove::teammate(4,agent);
		   */
		  if(nearest_opp.x > -10.0||wm.ball().pos().x > 0.0){
			  agent->debugClient().addMessage( "Pos(4)" );
			  return Vector2D(-40.0,0);
		  }
		  else
		  {
			  Vector2D point = nearest_opp - Vector2D(20.0,0);
			  if( point.x < -51.0){
				  point.x = -51.0;
			  }
			  Vector2D move_point = rcsc::Line2D::intersection(rcsc::Line2D( rcsc::Vector2D(point.x,-35.0),Vector2D(point.x,35.0)),
					  rcsc::Line2D( rcsc::Vector2D( -52.5, 0.0 ),wm.ball().pos()));
			  agent->debugClient().addMessage( "Pos(5)" );
			  return move_point;
		  }

	  }
}
/*-------------------------------------------------------------------*/
/*!

 */
double
Tactical_NAME_BasicMove_GoalieBasic::getBasicDashPower( PlayerAgent * agent,
		const Vector2D & move_point )
{
	const WorldModel & wm = agent->world();
	const PlayerType & mytype = wm.self().playerType();

	const double my_inc = mytype.staminaIncMax() * wm.self().recovery();

	if ( std::fabs( wm.self().pos().x - move_point.x ) > 1.0 )
	{
		return ServerParam::i().maxDashPower();
	}

	if ( wm.ball().pos().x > -30.0 )
	{
		if ( wm.self().stamina() < ServerParam::i().staminaMax() * 0.5 )
		{
			return my_inc * 0.5;
		}
		agent->debugClient().addMessage( "P1" );
		return my_inc;
	}
	else if ( wm.ball().pos().x > ServerParam::i().ourPenaltyAreaLineX() )
	{
		if ( wm.ball().pos().absY() > 20.0 )
		{
			// penalty area
			agent->debugClient().addMessage( "P2" );
			return my_inc;
		}
		if ( wm.ball().vel().x > 1.0 )
		{
			// ball is moving to opponent side
			agent->debugClient().addMessage( "P2.5" );
			return my_inc * 0.5;
		}

		int opp_min = wm.interceptTable()->opponentReachCycle();
		if ( opp_min <= 3 )
		{
			agent->debugClient().addMessage( "P2.3" );
			return ServerParam::i().maxDashPower();
		}

		/*        if ( wm.self().stamina() < ServerParam::i().staminaMax() * 0.7 )
        {
            agent->debugClient().addMessage( "P2.6" );
            return my_inc * 0.7;
        }
		 */
		agent->debugClient().addMessage( "P3" );
		return ServerParam::i().maxDashPower() * 0.3;
	}
	else
	{
		if ( wm.ball().pos().absY() < 15.0
				|| wm.ball().pos().y * wm.self().pos().y < 0.0 ) // opposite side
		{
			agent->debugClient().addMessage( "P4" );
			return ServerParam::i().maxDashPower();
		}
        else
        {
            agent->debugClient().addMessage( "P5" );
            return my_inc;
        }
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
Tactical_NAME_BasicMove_GoalieBasic::doPrepareDeepCross( PlayerAgent * agent,
                                         const Vector2D & move_point )
{
    if ( move_point.absY() < ServerParam::i().goalHalfWidth() - 0.8 )
    {
        // consider only very deep cross
        dlog.addText( Logger::TEAM,
                      __FILE__": doPrepareDeepCross no deep cross" );
        return false;
    }

    const WorldModel & wm = agent->world();

    const Vector2D goal_c( - ServerParam::i().pitchHalfLength(), 0.0 );

    Vector2D goal_to_ball = wm.ball().pos() - goal_c;

    if ( goal_to_ball.th().abs() < 60.0 )
    {
        // ball is not in side cross area
        dlog.addText( Logger::TEAM,
                      __FILE__": doPrepareDeepCross.ball is not in side cross area" );
        return false;
    }

    Vector2D my_inertia = wm.self().inertiaFinalPoint();
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;
    //double dist_thr = 0.5;

    if ( my_inertia.dist( move_point ) > dist_thr )
    {
        // needed to go to move target point
        double dash_power = getBasicDashPower( agent, move_point );
        dlog.addText( Logger::TEAM,
                      __FILE__": doPrepareDeepCross. need to move. power=%.1f",
                      dash_power );
        agent->debugClient().addMessage( "DeepCrossMove%.0f", dash_power );
        agent->debugClient().setTarget( move_point );
        agent->debugClient().addCircle( move_point, dist_thr );

        doGoToPointLookBall( agent,
                             move_point,
                             wm.ball().angleFromSelf(),
                             dist_thr,
                             dash_power );
        return true;
    }

    AngleDeg body_angle = ( wm.ball().pos().y < 0.0
    		? 10.0
    				: -10.0 );
    const double x_buf = 0.5;
    double dash_power = getBasicDashPower( agent, move_point );
    rcsc::Vector2D move_point_force = move_point;
    rcsc::Vector2D goal_1( -52.5 , ServerParam::i().goalHalfWidth() );
    rcsc::Vector2D goal_2( -52.5 , 0 );
    //ボールの位置、ゴールの中心、ゴールポストで三角形を2つ作る
    Triangle2D triangle_right = rcsc::Triangle2D( wm.ball().pos() , goal_1 , goal_2);
    Triangle2D triangle_left = rcsc::Triangle2D( wm.ball().pos() , goal_1 , goal_2);
    int mate_count_right=0, mate_count_left=0;

    const PlayerPtrCont::const_iterator mates_end = wm.teammatesFromSelf().end();
    for (PlayerPtrCont::const_iterator mate = wm.teammatesFromSelf().begin();
    		mate != mates_end; ++mate)
    {
    	if( triangle_right.contains( (*mate)->pos() )){
    		mate_count_right++;
    	}
    	else if( triangle_left.contains( (*mate)->pos() ) ){
    		mate_count_left++;
    	}
    }

    if( mate_count_right > mate_count_left  ){
    	move_point_force.y -= 2.0;
    }
    else if( mate_count_left > mate_count_right  ){
    	move_point_force.y += 2.0;
    }

    if ( ! Body_GoToPoint( move_point_force, x_buf, dash_power
                                      ).execute( agent ) )
               {
                   AngleDeg body_angle = ( wm.self().body().degree() > 0.0
                                           ? 90.0
                                           : -90.0 );
                   Body_TurnToAngle( body_angle ).execute( agent );

               }
    agent->debugClient().addMessage( "PrepareCross" );
    dlog.addText( Logger::TEAM,
    		__FILE__": doPrepareDeepCross  body angle = %.1f  move_point(%.1f %.1f)",
    		body_angle.degree(),
    		move_point.x, move_point.y );
    agent->debugClient().setTarget( move_point );

    agent->setNeckAction( new Neck_GoalieTurnNeck() );
    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
Tactical_NAME_BasicMove_GoalieBasic::doStopAtMovePoint( PlayerAgent * agent,
                                        const Vector2D & move_point )
{
    //----------------------------------------------------------
    // already exist at target point
    // but inertia movement is big
    // stop dash

    const WorldModel & wm = agent->world();
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    // now, in the target area
    if ( wm.self().pos().dist( move_point ) < dist_thr )
    {
        const Vector2D my_final
            = inertia_final_point( wm.self().pos(),
                                   wm.self().vel(),
                                   wm.self().playerType().playerDecay() );
        // after inertia move, can stay in the target area
        if ( my_final.dist( move_point ) < dist_thr )
        {
            agent->debugClient().addMessage( "InertiaStay" );
            dlog.addText( Logger::TEAM,
                          __FILE__": doStopAtMovePoint. inertia stay" );
            return false;
        }

        // try to stop at the current point
        dlog.addText( Logger::TEAM,
                      __FILE__": doStopAtMovePoint. stop dash" );
        agent->debugClient().addMessage( "Stop" );
        agent->debugClient().setTarget( move_point );

        Body_StopDash( true ).execute( agent ); // save recovery
        agent->setNeckAction( new Neck_GoalieTurnNeck() );
        return true;
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
Tactical_NAME_BasicMove_GoalieBasic::doMoveForDangerousState( PlayerAgent * agent,
                                              const Vector2D & move_point )
{
    const WorldModel& wm = agent->world();

    const double x_buf = 0.5;

    const Vector2D ball_next = wm.ball().pos() + wm.ball().vel();

    dlog.addText( Logger::TEAM,
                  __FILE__": doMoveForDangerousState" );

    if ( std::fabs( move_point.x - wm.self().pos().x ) > x_buf
         && ball_next.x < -ServerParam::i().pitchHalfLength() + 11.0
         && ball_next.absY() < ServerParam::i().goalHalfWidth() + 1.0 )
    {
        // x difference to the move point is over threshold
        // but ball is in very dangerous area (just front of our goal)

        // and, exist opponent close to ball
        if ( ! wm.opponentsFromBall().empty()
             && wm.opponentsFromBall().front()->distFromBall() < 2.0 )
        {
            Vector2D block_point
                = wm.opponentsFromBall().front()->pos();
            Line2D ball_line ( wm.ball().pos() , rcsc::Vector2D( -52.5, 0 ) );
            block_point.x -= 2.5;
            block_point.y = ball_line.getY( block_point.x );

            if ( wm.self().pos().x < block_point.x )
            {
                block_point.x = wm.self().pos().x;
            }

            dlog.addText( Logger::TEAM,
                          __FILE__": block opponent kickaer" );
            agent->debugClient().addMessage( "BlockOpp" );

            if ( doGoToMovePoint( agent, block_point ) )
            {
                return true;
            }

            double dist_thr = wm.ball().distFromSelf() * 0.1;
            if ( dist_thr < 0.5 ) dist_thr = 0.5;

            agent->debugClient().setTarget( block_point );
            agent->debugClient().addCircle( block_point, dist_thr );

            doGoToPointLookBall( agent,
                                 move_point,
                                 wm.ball().angleFromSelf(),
                                 dist_thr,
                                 ServerParam::i().maxDashPower() );
            return true;
        }
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
Tactical_NAME_BasicMove_GoalieBasic::doCorrectX( PlayerAgent * agent,
                                 const Vector2D & move_point )
{
    const WorldModel & wm = agent->world();

    const double x_buf = 0.5;

    dlog.addText( Logger::TEAM,
                  __FILE__": doCorrectX" );
    if ( std::fabs( move_point.x - wm.self().pos().x ) < x_buf )
    {
        // x difference is already small.
        dlog.addText( Logger::TEAM,
                      __FILE__": doCorrectX. x diff is small" );
        return false;
    }

    int opp_min_cyc = wm.interceptTable()->opponentReachCycle();
    if ( ( ! wm.existKickableOpponent() && opp_min_cyc >= 4 )
         || wm.ball().distFromSelf() > 18.0 )
    {
        double dash_power = getBasicDashPower( agent, move_point );

        dlog.addText( Logger::TEAM,
                      __FILE__": doCorrectX. power=%.1f",
                      dash_power );
        agent->debugClient().addMessage( "CorrectX%.0f", dash_power );
        agent->debugClient().setTarget( move_point );
        agent->debugClient().addCircle( move_point, x_buf );

        if ( ! wm.existKickableOpponent()
             && wm.ball().distFromSelf() > 30.0 )
        {
        	Line2D block_line( rcsc::Vector2D(-52.5 , 0) , wm.ball().pos() );
        	Vector2D move_point_force = move_point;
        	move_point_force.y = block_line.getY( move_point.x );
        	if( wm.ball().pos().x < -40.0 ){
        		move_point_force.x = -51.0;
        	}
        	agent->debugClient().addMessage( "move_point_force" );
        	agent->debugClient().setTarget( move_point_force );
            if ( ! Body_GoToPoint( move_point_force, x_buf, dash_power
                                   ).execute( agent ) )
            {
                AngleDeg body_angle = ( wm.self().body().degree() > 0.0
                                        ? 90.0
                                        : -90.0 );
                Body_TurnToAngle( body_angle ).execute( agent );

            }
            agent->setNeckAction( new Neck_TurnToBall() );
            return true;
        }

        doGoToPointLookBall( agent,
                             move_point,
                             wm.ball().angleFromSelf(),
                             x_buf,
                             dash_power );
        return true;
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
Tactical_NAME_BasicMove_GoalieBasic::doCorrectBodyDir( PlayerAgent * agent,
                                       const Vector2D & move_point,
                                       const bool consider_opp )
{
    // adjust only body direction

    const WorldModel & wm = agent->world();

    const Vector2D ball_next = wm.ball().pos() + wm.ball().vel();

    const AngleDeg target_angle = ( ball_next.y < 0.0 ? -90.0 : 90.0 );
    const double angle_diff = ( wm.self().body() - target_angle ).abs();

    dlog.addText( Logger::TEAM,
                  __FILE__": doCorrectBodyDir" );

    if ( angle_diff < 5.0 )
    {
        return false;
    }

#if 1
    {
        const Vector2D goal_c( - ServerParam::i().pitchHalfLength(), 0.0 );
        Vector2D goal_to_ball = wm.ball().pos() - goal_c;
        if ( goal_to_ball.th().abs() >= 60.0 )
        {
            dlog.addText( Logger::TEAM,
                          __FILE__": doCorrectBodyDir. danger area" );
            return false;
        }
    }
#else
    if ( wm.ball().pos().x < -36.0
         && wm.ball().pos().absY() < 15.0
         && wm.self().pos().dist( move_point ) > 1.5 )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": doCorrectBodyDir. danger area" );
        return false;
    }
#endif

    double opp_ball_dist
        = ( wm.opponentsFromBall().empty()
            ? 100.0
            : wm.opponentsFromBall().front()->distFromBall() );
    if ( ! consider_opp
         || opp_ball_dist > 7.0
         || wm.ball().distFromSelf() > 20.0
         || ( std::fabs( move_point.y - wm.self().pos().y ) < 1.0 // y diff
              && ! wm.existKickableOpponent() ) )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": body face to %.1f.  angle_diff=%.1f %s",
                      target_angle.degree(), angle_diff,
                      consider_opp ? "consider_opp" : "" );
        agent->debugClient().addMessage( "CorrectBody%s",
                                         consider_opp ? "WithOpp" : "" );
        Body_TurnToAngle( target_angle ).execute( agent );
        agent->setNeckAction( new Neck_GoalieTurnNeck() );
        return true;
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
Tactical_NAME_BasicMove_GoalieBasic::doGoToMovePoint( PlayerAgent * agent,
                                      const Vector2D & move_point )
{
    // move to target point
    // check Y coordinate difference

    const WorldModel & wm = agent->world();

    double dist_thr = wm.ball().distFromSelf() * 0.08;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    const double y_diff = std::fabs( move_point.y - wm.self().pos().y );
    if ( y_diff < dist_thr )
    {
        // already there
        dlog.addText( Logger::TEAM,
                      __FILE__": doGoToMovePoint. y_diff=%.2f < thr=%.2f",
                      y_diff, dist_thr );
        return false;
    }

    //----------------------------------------------------------//
    // dash to body direction

    double dash_power = getBasicDashPower( agent, move_point );

    // body direction is OK
    if ( std::fabs( wm.self().body().abs() - 90.0 ) < 7.0 )
    {
        // calc dash power only to reach the target point
        double required_power = y_diff / wm.self().dashRate();
        if ( dash_power > required_power )
        {
            dash_power = required_power;
        }

        if ( move_point.y > wm.self().pos().y )
        {
            if ( wm.self().body().degree() < 0.0 )
            {
                dash_power *= -1.0;
            }
        }
        else
        {
            if ( wm.self().body().degree() > 0.0 )
            {
                dash_power *= -1.0;
            }
        }

        dash_power = ServerParam::i().normalizeDashPower( dash_power );

        dlog.addText( Logger::TEAM,
                      __FILE__": doGoToMovePoint. CorrectY(1) power= %.1f",
                      dash_power );
        agent->debugClient().addMessage( "CorrectY(1)%.0f", dash_power );
        agent->debugClient().setTarget( move_point );

        agent->doDash( dash_power );
        agent->setNeckAction( new Neck_GoalieTurnNeck() );
    }
    else
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": doGoToMovePoint. CorrectPos power= %.1f",
                      dash_power );
        agent->debugClient().addMessage( "CorrectPos%.0f", dash_power );
        agent->debugClient().setTarget( move_point );
        agent->debugClient().addCircle( move_point, dist_thr );

        doGoToPointLookBall( agent,
                             move_point,
                             wm.ball().angleFromSelf(),
                             dist_thr,
                             dash_power );
    }
    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
Tactical_NAME_BasicMove_GoalieBasic::doGoToPointLookBall( PlayerAgent * agent,
                                          const Vector2D & target_point,
                                          const AngleDeg & body_angle,
                                          const double & dist_thr,
                                          const double & dash_power,
                                          const double & back_power_rate )
{
    const WorldModel & wm = agent->world();

    if ( wm.gameMode().type() == GameMode::PlayOn
         || wm.gameMode().type() == GameMode::PenaltyTaken_ )
    {
        agent->debugClient().addMessage( "Goalie:GoToLook" );
        dlog.addText( Logger::TEAM,
                      __FILE__": doGoToPointLookBall. use GoToPointLookBall" );
        Bhv_GoToPointLookBall( target_point,
                               dist_thr,
                               dash_power,
                               back_power_rate
                               ).execute( agent );
    }
    else
    {
        agent->debugClient().addMessage( "Goalie:GoTo" );
        dlog.addText( Logger::TEAM,
                      __FILE__": doGoToPointLookBall. use GoToPoint" );
        if ( Body_GoToPoint( target_point, dist_thr, dash_power
                             ).execute( agent ) )
        {
            dlog.addText( Logger::TEAM,
                          __FILE__": doGoToPointLookBall. go" );
        }
        else
        {
            Body_TurnToAngle( body_angle ).execute( agent );
            dlog.addText( Logger::TEAM,
                          __FILE__": doGoToPointLookBall. turn to %.1f",
                          body_angle.degree() );
        }

        agent->setNeckAction( new Neck_TurnToBall() );
    }
}


const
PlayerObject* Tactical_NAME_BasicMove_GoalieBasic::teammate( const int unum ,const rcsc::PlayerAgent * agent)
{
	const rcsc::WorldModel & wm = agent->world();
    const rcsc::PlayerPtrCont & mates = wm.teammatesFromSelf();
    const rcsc::PlayerPtrCont::const_iterator mate_begin = mates.begin();
    const rcsc::PlayerPtrCont::const_iterator mates_end = mates.end();
      for( rcsc::PlayerPtrCont::const_iterator p = mates.begin();
               p != mates_end;
               ++p )
    {
        if ( (*p)->unum() == unum )
        {
            return *p;
        }
    }

    return *mate_begin;
}


